// Print help message for the specified command.
#ifndef FISH_PRINT_HELP_H
#define FISH_PRINT_HELP_H

/// Print help message for the specified command.
void print_help(const char *cmd, int fd);

#endif
